/*
  clk_drv.h

  Copyright(C) 2001 Yamaha Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  02110-1301, USA

  Description:
    NbN}l[W[̊O錾t@C
 */
#ifndef _CLK_DRV_
#define _CLK_DRV_

#include "ymh_type.h"

#ifdef _DEBUG
#define CLKDRV_DEBUG
#endif /* _DEBUG */

/* [gNbNɃNbN\[XATCĂȂ */
#define CLOCK_SOURCE_NOT_ASSIGNED (-1)

/* NbNIuWFNgɏo͐ݒ肪ĂȂ */
#define OUTPUT_TYPE_NOT_ASSIGNED  (-1)

/* NbNIuWFNg̕\̏l */
#define CLOCK_DEFAULT_DIVISION    (480)

/* NbNIuWFNg̃e|̏l */
#define CLOCK_DEFAULT_TEMPO       (500000)

/* [gNbN̑΃e|̏l */
#define CLOCK_DEFAULT_REL_TEMPO   (100000)

/* vZXNbNIuWFNg~Ƃ̃NbNl */
#define CLOCK_STOP_FROM_PROCESS   (-1)

/* NbN\[X̃foCXԍ */
enum {
	CLOCK_DEVICE_SERIAL_0	/* Serial|[gOAK2Xł͖gp */
	, CLOCK_DEVICE_SERIAL_1	/* Serial|[gP */
	, CLOCK_DEVICE_USB		/* USB|[g */
	, CLOCK_DEVICE_MLAN		/* mLan|[g */
	, CLOCK_DEVICE_NETWORK	/* Network|[g */
	, CLOCK_DEVICE_SUM		/* foCXԍ̍ől */
	, CLOCK_DEVICE_NOT_ASSIGNED = CLOCK_DEVICE_SUM
};

/* NbN\[X^Cv */
enum {
	CLOCK_SOURCE_TYPE_INTERNAL
	, CLOCK_SOURCE_TYPE_AUTO_CLOCK
	, CLOCK_SOURCE_TYPE_MIDI_CLOCK
	, CLOCK_SOURCE_TYPE_MTC_CLOCK
	, CLOCK_SOURCE_TYPE_SUM
};

/* o̓^Cv */
enum {
	CLOCK_OUTPUT_TYPE_MID
	, CLOCK_OUTPUT_TYPE_MTC
	, CLOCK_OUTPUT_TYPE_RTM
	, CLOCK_OUTPUT_TYPE_SUM
};

/* MIDISongContorolǂ */
enum {
	CLOCK_DISABLE_SONG_CONTROL
	, CLOCK_ENABLE_SONG_CONTROL
};

/* oSMPTE^Cv */
#define CLOCK_SMPTE_24_FRAME           (0)  /* 24 frames per second                  */
#define CLOCK_SMPTE_25_FRAME           (1)  /* 25 frames per second                  */
#define CLOCK_SMPTE_30_DROP_FRAME      (2)  /* 30 frames per second (drop frame)     */
#define CLOCK_SMPTE_30_NON_DROP_FRAME  (3)  /* 30 frames per second (non drop frame) */
#define CLOCK_SMPTE_OTHER_FRAME        (4)  /* other frames per second               */
#define CLOCK_SMPTE_TYPE_MAX           CLOCK_SMPTE_OTHER_FRAME

/* ʃbZ[W */
#define CLOCK_MSG_ID_BIT_SHIFT     (0)
#define CLOCK_MSG_SIZE_BIT_SHIFT   (16)
#define CLOCK_MSG_NUM_BIT_SHIFT    (24)
#define CLOCK_MASK_MSG_ID_BIT      (0x0000FFFF)
#define CLOCK_MASK_MSG_SIZE_BIT    (0x00FF0000)
#define CLOCK_MASK_MSG_NUM_BIT     (0xFF000000)

#define CLOCK_MESSAGE_EXIT         (CLOCK_MASK_MSG_ID_BIT)
#define CLOCK_MESSAGE_NON          (0x0000)
#define CLOCK_MESSAGE_REWIND_START (0x0001) /* 0xFA */
#define CLOCK_MESSAGE_START        (0x0002) /* 0xFB */
#define CLOCK_MESSAGE_STOP         (0x0003) /* 0xFC */
#define CLOCK_MESSAGE_SONG_POS     (0x0004|(2<<CLOCK_MSG_SIZE_BIT_SHIFT)|(1<<CLOCK_MSG_NUM_BIT_SHIFT)) /* 0xF2 */
#define CLOCK_MESSAGE_MTC_FULL     (0x0005|(1<<CLOCK_MSG_SIZE_BIT_SHIFT)|(5<<CLOCK_MSG_NUM_BIT_SHIFT)) /* 0xF0 */
#define CLOCK_MESSAGE_PAUSE_ON     (0x0006)
#define CLOCK_MESSAGE_PAUSE_OFF    (0x0007)

/* G[R[h */
#define E_OK (0)
#define E_ERR (-1)
#define E_NOMEM (-2)
#define E_BADOBJ (-3)


/* -------------------------------------- */
/*  f[Vݒ胁bZ[W̍\  */
/* -------------------------------------- */
typedef struct _clock_delay_t {
	SINT32 duration;			/* ̃Cxg܂ł̃f[V */
	UINT32 header;				/* bZ[W̃wb_ */
	void* data;					/* bZ[Wϒւ̃|C^ */
} clock_delay_t;

/* -------------------------- */
/*  ϒbZ[W̍\  */
/* -------------------------- */
typedef struct _clock_message_t {
	UINT32 header;				/* bZ[W̃wb_ */
    /* oCg  FMbZ[W               */
    /* oCg  FbZ[WTCY(0A1A2A4)   */
    /* OloCg  FbZ[WID                   */
	void* data;					/* bZ[Wϒւ̃|C^ */
} clock_message_t;

/* ----------------------------- */
/*  SMPTEݒ胁bZ[W̍\  */
/* ----------------------------- */
typedef struct _clock_smpte_t {
	int		type;				/* SMPTẼ^Cv */
	UINT32	division;			/* SMPTE̕\ */
} clock_smpte_t;

/* -------------------------------------------------- */
/*  NbNIuWFNgɃA^b`vZX̏  */
/* -------------------------------------------------- */
typedef struct _clock_info_t {
	char* name;					/* O */
	int nameSize;				/* O̕ */
	char* info;					/*  */
	int infoSize;				/* ̕ */
} clock_info_t;

/* -------------------------- */
/* t[ϐ̃CfbNX */
/* -------------------------- */
enum {
	Index_Type,
	Index_Hours,
	Index_Minutes,
	Index_Seconds,
	Index_Frames,
	MtcIndex_Sum
};

/* ------------------------- */
/*  ioctlp[^̍\  */
/* ------------------------- */
typedef struct _clock_parameter_t {
	void* objectPointer;		/* kernelԂłclock objectւ̃|C^ */
	union parameter_t {
		clock_delay_t	delay;	/* CLOCK_IOCTL_GET_MESSAGE͂ */
		clock_message_t message;
		clock_smpte_t	smpte;
		clock_info_t	info;
		SINT32			clock;
		int				output;
		UINT32			tempo;
		UINT32			division;
		int				flag;
		SINT8			frame[MtcIndex_Sum];
	} u;						/* p[^ */
} clock_parameter_t;


#if defined(CLK_IF_LIBRARY) || defined (__KERNEL__)

/* ȉclk_Delay̖߂l */
#define CLOCK_NO_MESSAGE     (0x00)
#define CLOCK_EXIST_TIMING   (0x01<<0)
#define CLOCK_EXIST_MESSAGE  (0x01<<1)
#define CLOCK_ADVANCE_CLOCK  (0x01<<2)

/* ȉioctl̑ */
#define DRV_MAGIC 'c'
#define CLOCK_IOCTL_SET_CLOCK_SOURCE        _IO(DRV_MAGIC, 1)
#define CLOCK_IOCTL_GET_CLOCK_SOURCE        _IO(DRV_MAGIC, 2)
#define CLOCK_IOCTL_SET_CLOCK_DEVICE        _IO(DRV_MAGIC, 3)
#define CLOCK_IOCTL_GET_CLOCK_DEVICE        _IO(DRV_MAGIC, 4)
#define CLOCK_IOCTL_SET_DEVICE_SONG_CONTROL _IO(DRV_MAGIC, 5)
#define CLOCK_IOCTL_GET_DEVICE_SONG_CONTROL _IO(DRV_MAGIC, 6)
#define CLOCK_IOCTL_SET_MTC_OFFSET          _IO(DRV_MAGIC, 7)
#define CLOCK_IOCTL_GET_MTC_OFFSET          _IO(DRV_MAGIC, 8)
#define CLOCK_IOCTL_START_CLOCK             _IO(DRV_MAGIC, 9)
#define CLOCK_IOCTL_STOP_CLOCK              _IO(DRV_MAGIC, 10)
#define CLOCK_IOCTL_ADVANCE_CLOCK           _IO(DRV_MAGIC, 11)
#define CLOCK_IOCTL_BROADCAST_MESSAGE       _IO(DRV_MAGIC, 12)
#define CLOCK_IOCTL_SET_RELATIVE_TEMPO      _IO(DRV_MAGIC, 13)
#define CLOCK_IOCTL_GET_RELATIVE_TEMPO      _IO(DRV_MAGIC, 14)
#define CLOCK_IOCTL_RESET_CLOCK             _IO(DRV_MAGIC, 15)
#define CLOCK_IOCTL_GET_RUN_STATUS          _IO(DRV_MAGIC, 16)
#define CLOCK_IOCTL_GET_OBJECT_SUM          _IO(DRV_MAGIC, 17)
#define CLOCK_IOCTL_GET_OBJECT_POINTER      _IO(DRV_MAGIC, 18)
#define CLOCK_IOCTL_GET_ALL_APP_INFO_SIZE   _IO(DRV_MAGIC, 19)
#define CLOCK_IOCTL_GET_ALL_APP_INFO        _IO(DRV_MAGIC, 20)
#define CLOCK_IOCTL_INITIALIZE_ROOT_CLOCK   _IO(DRV_MAGIC, 21)
#define CLOCK_IOCTL_ATTACH_OBJECT           _IO(DRV_MAGIC, 22)
#define CLOCK_IOCTL_DETACH_OBJECT           _IO(DRV_MAGIC, 23)
#define CLOCK_IOCTL_SEND_MESSAGE            _IO(DRV_MAGIC, 24)
#define CLOCK_IOCTL_START_OBJECT            _IO(DRV_MAGIC, 25)
#define CLOCK_IOCTL_STOP_OBJECT             _IO(DRV_MAGIC, 26)
#define CLOCK_IOCTL_DELAY                   _IO(DRV_MAGIC, 27)
#define CLOCK_IOCTL_SET_MIDI_SONG_CONTROL   _IO(DRV_MAGIC, 28)
#define CLOCK_IOCTL_GET_MIDI_SONG_CONTROL   _IO(DRV_MAGIC, 29)
#define CLOCK_IOCTL_PLUS_CLOCK              _IO(DRV_MAGIC, 30)
#define CLOCK_IOCTL_SET_CLOCK               _IO(DRV_MAGIC, 31)
#define CLOCK_IOCTL_GET_CLOCK               _IO(DRV_MAGIC, 32)
#define CLOCK_IOCTL_SET_DIV                 _IO(DRV_MAGIC, 33)
#define CLOCK_IOCTL_GET_DIV                 _IO(DRV_MAGIC, 34)
#define CLOCK_IOCTL_SET_TMP                 _IO(DRV_MAGIC, 35)
#define CLOCK_IOCTL_GET_TMP                 _IO(DRV_MAGIC, 36)
#define CLOCK_IOCTL_SET_OUTPUT_TYPE         _IO(DRV_MAGIC, 37)
#define CLOCK_IOCTL_GET_OUTPUT_TYPE         _IO(DRV_MAGIC, 38)
#define CLOCK_IOCTL_SET_SMPTE               _IO(DRV_MAGIC, 39)
#define CLOCK_IOCTL_GET_SMPTE               _IO(DRV_MAGIC, 40)
#define CLOCK_IOCTL_GET_OBJECT_RUN_STATUS   _IO(DRV_MAGIC, 41)
#define CLOCK_IOCTL_SET_APP_INFO            _IO(DRV_MAGIC, 42)
#define CLOCK_IOCTL_GET_MESSAGE             _IO(DRV_MAGIC, 43)
#define CLOCK_IOCTL_RESTART_CLOCK           _IO(DRV_MAGIC, 44)

#define DRV_MAXNR 44

#endif /* CLK_MNG_LIBRARY || __KERNEL__ */

#ifdef __KERNEL__

#include "ngs_drv.h"
#if	LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
#include <linux/tqueue.h>
#endif

#define CLKDRV_LOCK   { unsigned long flags; \
                        save_flags(flags); \
                        cli();
#define CLKDRV_UNLOCK restore_flags(flags); }


/* ------------------------ */
/* Map Func Object ֐ */
/* ------------------------ */
typedef struct {
	ERR (*createMapFuncObj)(void** ppMapFuncObj, int output);               /* MapFuncObject֐   */
	ERR (*destroyMapFuncObj)(void** ppMapFuncObj);                          /* MapFuncObjectj֐   */
	ERR (*startMapFuncObj)(void* pMapFuncObj);                              /* X^[g֐            */
	ERR (*stopMapFuncObj)(void* pMapFuncObj, UINT32* pClock);               /* Xgbv֐            */
	ERR (*resetMapFuncObj)(void* pMapFuncObj);                              /* Zbg֐            */
	ERR (*setTempo)(void* pMapFuncObj, UINT32 tempo, SINT32 clock);         /* e|ύX֐          */
	ERR (*getTempo)(void* pMapFuncObj, UINT32* pTempo);                     /* e|擾֐          */
	ERR (*setDivision)(void* pMapFuncObj, UINT32 division);                 /* \ύX֐          */
	ERR (*getDivision)(void* pMapFuncObj, UINT32* pDivision);               /* \擾֐          */
	ERR (*setRelativeTempo)(void* pMapFuncObj, UINT32 tempo);               /* ΃e|ύX֐      */
	ERR (*getRelativeTempo)(void* pMapFuncObj, UINT32* pTempo);             /* ΃e|擾֐      */
	UINT32 (*convertAdv)(void* pMapFuncObj, UINT32 nTime);                  /* Advance ClockԎϊ */
	UINT32 (*convertInt)(void* pMapFuncObj, UINT32 nTime, SINT32* thrs);    /* InternalClockԎϊ */
	UINT32 (*convertExt)(void* pMapFuncObj, void* pTrgData, SINT32* thrs);  /* ExternalClockԎϊ */
} CLK_MAP_OBJ_FUNC;

/* ----------------------------- */
/* Clock Source Manager ֐ */
/* ----------------------------- */
typedef struct {
	ERR (*registerRootClock)(void* pRootClock, const int sourceType, const CLK_MAP_OBJ_FUNC** pptMapObjFunc);
																			/* RootClocko^֐       */
	ERR (*deregisterRootClock)(void* pRootClock, const int sourceType);		/* RootClock֐       */
	ERR (*startClock)(void);												/* X^[g֐            */
	ERR (*stopClock)(void);													/* Xgbv֐            */
} CLK_SRC_MNG_FUNC;

/* ----------------------------- */
/* Clock Device Manager ֐ */
/* ----------------------------- */
typedef struct {
	ERR (*registerRootClock)(void* pRootClock, int deviceNum);				/* RootClocko^֐    */
	ERR (*deregisterRootClock)(void* pRootClock, int deviceNum);			/* RootClock֐    */
	ERR (*initializeMtcCounter)(void* pRootClock, int deviceNum);			/* JE^֐	*/
	ERR (*setMtcOffsetCounter)(void* pRootClock, int deviceNum, SINT8* frame_data);	/* ItZbgݒ֐  */
	ERR (*getMtcOffsetCounter)(void* pRootClock, int deviceNum, SINT8* frame_data);	/* ItZbg擾֐  */
	ERR (*setAutoMidiClockTempo)(void* pRootClock, int deviceNum, UINT32 tempo);    /* Auto Tempoݒ֐  */
} CLK_DEV_MNG_FUNC;

extern void clkdrv_setClockSourceFunc(const CLK_SRC_MNG_FUNC* ptClkSrcFunc);
extern void clkdrv_clearClockSourceFunc(void);
extern void clkdrv_setClockDeviceFunc(const CLK_DEV_MNG_FUNC* ptClkDevFunc);
extern void clkdrv_clearClockDeviceFunc(void);

extern ERR clkdrv_receiveStartMessage(void* pRootClock, const UINT32 message);
extern ERR clkdrv_receiveStopMessage(void* pRootClock, const int sourceType);
extern ERR clkdrv_receiveSongPositionMessage(void* pRootClock, const UINT16 pos);
extern ERR clkdrv_receiveMtcFullMessage(void* pRootClock, SINT8* frame_data);
extern ERR clkdrv_advanceClockInt(void* pRootClock, UINT32 time);
extern ERR clkdrv_advanceClockExt(void* pRootClock, const int sourceType, void* pTrgData);

/* ---------------------------------- */
/* Clock Source ManagerI֐ */
/* ---------------------------------- */
extern int clksrc_init(void);
extern void clksrc_cleanup(void);

/* ---------------------------------- */
/* Clock Device ManagerI֐ */
/* ---------------------------------- */
extern int clkdev_init(void);
extern void clkdev_cleanup(void);

/* ----------------------------- */
/* Clock Device R[obN֐ */
/* ----------------------------- */
extern void clkdev_receiveSerialMidiData(int line, unsigned char ch);
extern void clkdev_receiveUsbMidiData(int line, unsigned char ch);

#endif /* __KERNEL__ */

#endif /* _CLK_DRV_ */
