/*
  clk_drv.c

  Copyright(C) 1993 Yamaha Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  02110-1301, USA

  Description:

  Revision history:
  Fri Mar 10 2006 : BuiltiñVAhCo㏑@~߂ăR[obNɂ
  Mon Jan 24 2011 : [U[ԂMIDI Data^悤ɏC
 */


#define __KMC_MODULE_DEBUG
#define __KMC_MODULE_NAME "/../../clk-drv/src/clk-serial.ko"

#include "clk_drv.h"

void serial_txx9_register_callbak(int line,void (*send)(int line,unsigned char ch));
void serial_txx9_unregister_callbak(int line);

/*==============================================
  internal table and ram declaration
==============================================*/
#ifdef CLKDRV_DEBUG
static const char* dmsg = "clk_serial";
#endif
static const char* devname = "clk_dev";

typedef struct {
	int line;			/* foCX}Ci[ԍ */
} clk_serial_t;

typedef struct {
	int nMajorNumber;	/* foCXW[ԍ */
	clk_serial_t device[ CLOCK_DEVICE_SUM ];
} clock_serial_manager_t;

static clock_serial_manager_t clockSerial;

static void clk_serial_callback( int line, unsigned char ch )
{
	clkdev_receiveSerialMidiData( line, ch );
}

static int clk_serial_Open(struct inode* pInode,  struct file* pFile)
{
	int nDeviceNum = MINOR(pInode->i_rdev);

	if( nDeviceNum >= CLOCK_DEVICE_SUM ){
#ifdef CLKDRV_DEBUG
		printk(KERN_DEBUG "%s: error clk_serial_Open \n", dmsg);
#endif
		return -ENODEV;
	}

	pFile->private_data = &(clockSerial.device[nDeviceNum]);

#if	LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
	MOD_INC_USE_COUNT;
#endif
	return E_OK;
}

static int clk_serial_Close(struct inode* pInode, struct file* pFile)
{
#if	LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
	MOD_DEC_USE_COUNT;
#endif
	return E_OK;
}

static ssize_t clk_serial_Read(struct file* pFile, char* pBuff, size_t count, loff_t* loff)
{
	return E_OK;
}

static ssize_t clk_serial_Write(struct file* pFile, const char* pBuff, size_t count, loff_t* loff)
{
	clk_serial_t* dev = pFile->private_data;
	size_t i;
	for( i = 0; i < count; ++i ){
		clkdev_receiveSerialMidiData( dev->line, *pBuff++ );
	}
#ifdef CLKDRV_DEBUG
	printk(KERN_DEBUG "%s: clk_serial_Write line = %d count = %d\n", dmsg, dev->line, count);
#endif
	return count;
}

/* -------------------------- */
/* file operation֐e[u */
/* -------------------------- */
static struct file_operations fileop = {
	owner:		NULL,            /* struct module *owner                                                    */
	llseek:		NULL,            /* loff_t (*llseek)(struct file*, loff_t, int);                            */
	read:		clk_serial_Read, /* ssize_t (*read)(struct file*, char*, size_t, loff_t*);                  */
	write:		clk_serial_Write,/* ssize_t (*write)(struct file*, const char*, size_t, loff_t*);           */
	readdir:	NULL,            /* int (*readdir)(struct file*, void*, filldir_t);                         */
	poll:		NULL,            /* unsigned int (*poll)(struct file*, struct poll_table_struct*);          */
	ioctl:		NULL,            /* int (*ioctl)(struct inode*, struct file*, unsigned int, unsigned long); */
	mmap:		NULL,            /* int (*mmap)(struct file*, struct vm_area_struct*);                      */
	open:		clk_serial_Open, /* int (*open)(struct inode*, struct file*);                               */
	flush:		NULL,            /* int (*flush)(struct file*);                                             */
	release:	clk_serial_Close,/* int (*release)(struct inode*, struct file*);                            */
	fsync:		NULL,            /* int (*fsync)(struct file*, struct dentry*, int datasync);               */
	fasync:		NULL,            /* int (*fasync)(int, struct file*, int);                                  */
	lock:		NULL,            /* int (*lock)(struct file*, int, struct file_lock*);                      */
	readv:		NULL,            /* ssize_t (*readv)(struct file*, const struct iovec*, unsigned long, loff_t*);  */
	writev:		NULL,            /* ssize_t (*writev)(struct file*, const struct iovec*, unsigned long, loff_t*); */
};

static int __init clk_serial_init(void)
{
	int nErr;
	int nCnt;

	/* NbN}l[WEhCo */
#ifndef AUTO_MAJOR_NUMBER
	clockSerial.nMajorNumber = 122;
#else
	clockSerial.nMajorNumber = 0;        /* 0 = W[ԍJ[l擾 */
#endif
	if( (nErr = register_chrdev( clockSerial.nMajorNumber, devname, &fileop )) < 0 ){
#ifdef CLKDRV_DEBUG
		if( nErr == -EBUSY ){
			printk(KERN_DEBUG "%s: error already in use. \n", dmsg);
		}
		else{
			printk(KERN_DEBUG "%s: error register_chrdev() err=%d \n", dmsg, nErr);
		}
#endif
	}

#ifdef AUTO_MAJOR_NUMBER
	clockSerial.nMajorNumber = nErr;
#endif

	for( nCnt = 0; nCnt < CLOCK_DEVICE_SUM; ++nCnt ){
		clockSerial.device[nCnt].line = nCnt;
	}

	serial_txx9_register_callbak( 1, clk_serial_callback );	/* Serial */
	serial_txx9_register_callbak( 2, clk_serial_callback );	/* USB */
	serial_txx9_register_callbak( 3, clk_serial_callback );	/* mLan */

#ifdef CLKDRV_DEBUG
	printk(KERN_DEBUG "%s: init_module()\n", dmsg);
#endif

	return E_OK;
}

static void __exit clk_serial_exit(void)
{
	serial_txx9_unregister_callbak( 1 );
	serial_txx9_unregister_callbak( 2 );
	serial_txx9_unregister_callbak( 3 );

	unregister_chrdev(clockSerial.nMajorNumber, devname);

#ifdef CLKDRV_DEBUG
	printk(KERN_DEBUG "%s: cleanup_module()\n", dmsg);
#endif
}

module_init(clk_serial_init);
module_exit(clk_serial_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Yamaha clock manager serial driver");
